/* @i
 * File: constructTree.c
 *
 * Form: C
 * Path: ~/Dewey/D5/D51/D511/D511.5_Graphs/Balance/Code/
 * Note: construct partially/full im/balanced trees
 * Meta: Phylogenetic Trees, Balanced, Chained
 * Date: 2011-06-03
 */

/* @ include files */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* @ definitions */
enum {FAILURE = -1, OK};
enum {MAXNAME = 100, MAXLINE = 1000, MAXLEN = 10000};
enum {BASE10 = 10};
enum {UNITY = 1, BASE2 = 2, PAIR = 2};
enum {UNBAL, BALAN};
#define SOME_DISTANCE 0.12345

/* @globals */
int iNodeWritten_GLOBAL = 0;
const int zzciUBB[][PAIR] = {{1,1}, {1,2}, {2,2}, {2,4}, {4,4}, {4,8}, {8,8}, 
// ***kb-13/6/12			     {8,16}, {16,16}, {16,32}, {32,32}};
			    {8,16}, {16,16}, {16,32}, {32,32}, {32,64}, {64,64}, 
				{64,128}, {128,128}, {128,256}, {256,256}};
// ***end of kb-13/6/12

/* @ prototypes */
int Commandline(int *iDepth_p, char zcFilename[], 
		int argc, char *argv[], char zcError[]);
void ErrorExit(char zcError[]);
int Ipow(int *iNodes_p, int iDepth, char zcError[]);
int CreateBalanced(int iD, int iDepth, FILE *pfOut, char **ppcSeqLabels, 
		   char zcAux[], char zcError[]);
int OpenFiles(FILE **pfBal_p, FILE **pfImb_p, FILE **pfIm2_p, FILE **pfIbb_p, 
	      char zcFilename[], 
	      char zcAux[], char zcError[]);
int ReadSequenceLabels(char ***ppcSeqLabels_p, char zcFilename[], int iNodes, 
		       char zcAux[], char zcError[]);
int CreateUnBalanced(int iD, int iNodes, FILE *pfOut, char **ppcSeqLabels, 
		     char zcAux[], char zcError[]);
int CreateTwoUnBalanced(int iD, int iNodes, FILE *pfOut, char **ppcSeqLabels, 
			char zcAux[], char zcError[]);
int CreateUnBalancedBalanced(int iD, int iDepth, int iNodes, FILE *pfOut, 
			     char **ppcSeqLabels,
			     char zcAux[], char zcError[]);

/* @* ilog2 */
int ilog2(int iNum){

  double dNum = (double)iNum;
  int i;

  for (i = 0; dNum >= BASE2; dNum/=BASE2, i++);

  return i;

} /* this is the end of ilog2() */



/* @* main */
int main (int argc, char *argv[]){

  /* @<variables local to main@> */
  int iDepth = 0; /* depth to which to build tree */
  int iNodes = 0; /* number of nodes in tree of depth iDepth */
  int iD = 0; /* current depth (distance from root) */
  char zcAux[MAXLEN] = {0}; /* auxilliary print string */
  char zcError[MAXLEN] = {0}; /* error message */
  char zcFilename[MAXNAME] = {0}; /* name of input file */
  FILE *pfBal = NULL;
  FILE *pfImb = NULL;
  FILE *pfIm2 = NULL;
  FILE *pfIbb = NULL;
  char **ppcSeqLabels = NULL;


  /* @<commandline@> */
  if (OK != Commandline(&iDepth, zcFilename, 
			argc, argv, zcError)){
    ErrorExit(zcError);
  }
  printf("%s:%s:%d: depth = %d\n", __FUNCTION__, __FILE__, __LINE__, iDepth);

  if (OK != Ipow(&iNodes, iDepth, zcError)){
    ErrorExit(zcError);
  }
  printf("%s:%s:%d: nodes = %d\n", __FUNCTION__, __FILE__, __LINE__, iNodes);

  /* @<read sequence labels> */
  if (OK != ReadSequenceLabels(&ppcSeqLabels, zcFilename, iNodes, 
			       zcAux, zcError)){
    ErrorExit(zcError);
  }


  /* @<open files for output> */
  if (OK != OpenFiles(&pfBal, &pfImb, &pfIm2, &pfIbb, 
		      zcFilename, zcAux, zcError)){
    ErrorExit(zcError);
  }

  /*----- @<recursive node> -----*/
  /* @<balanced tree> */
  iD = 1;
  iNodeWritten_GLOBAL = 0;
  if (OK != CreateBalanced(iD, iDepth, pfBal, ppcSeqLabels, zcAux, zcError)){
    ErrorExit(zcError);
  } 
  fprintf(pfBal, ";\n");

  /* @<unbalanced tree> */
  iD = 1;
  iNodeWritten_GLOBAL = 0;
  if (OK != CreateUnBalanced(iD, iNodes, pfImb, ppcSeqLabels, zcAux, zcError)){
    ErrorExit(zcError);
  } 

  /* @<2 unbalanced trees> */
  iD = 1;
  iNodeWritten_GLOBAL = 0;
  if (OK != CreateTwoUnBalanced(iD, iNodes, pfIm2, ppcSeqLabels, 
				zcAux, zcError)){
    ErrorExit(zcError);
  }
  fprintf(pfImb, ";\n");

#if 1
  /* @i NOTE: this does not work */
  /* @<unbalanced combination of balanced sub-trees> */  
  if (iDepth > sizeof(zzciUBB)/(PAIR*sizeof(int))-1){
    sprintf(zcError, 
	    "%s:%s:%d: cant go beyond level %ld, increase zzciUBB, recompile\n",
	    __FUNCTION__, __FILE__, __LINE__, 
	    sizeof(zzciUBB)/(PAIR*sizeof(int))-1);
    ErrorExit(zcError);
  }
  iNodeWritten_GLOBAL = 0;
  iD = 1;
  iNodes = zzciUBB[iDepth][UNBAL];
  iDepth = zzciUBB[iDepth][BALAN];
  if (OK != CreateUnBalancedBalanced(iD, iDepth, iNodes, pfIbb, ppcSeqLabels,
                                zcAux, zcError)){
    ErrorExit(zcError);
  }
  fprintf(pfIbb, ";\n");
#endif 

  /* @i 
   * prepare for exit */
  printf("+---------------------------+\n"
	 "| Program ran to completion |\n"
	 "+---------------------------+\n");
  /* @= */ 
  return OK;

} /* this is the end of main() */

/********************* EO MAIN ***********************************/

/* @* ReadSequenceLabels */
int ReadSequenceLabels(char ***ppcSeqLabels_p, char zcFilename[], int iNodes, 
		       char zcAux[], char zcError[]){

  FILE *pfIn = NULL; /* input file */
  char zcScanline[MAXLINE] = {0};
  int iL; 

  /* @<open file with seq labels for reading> */
  if (NULL == (pfIn = fopen(zcFilename, "r"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for reading\n", 
	    __FUNCTION__, __FILE__, __LINE__, zcFilename);
    return FAILURE;
  }


  /* @<allocate primary memory for sequence labels> */
  if (NULL == (*ppcSeqLabels_p = (char **)malloc(iNodes * sizeof(char **)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d char** for seq labels\n",
	    __FUNCTION__, __FILE__, __LINE__, iNodes);
    return FAILURE;
  }

  /* @<read file line by line> */
  iL = 0;
  while ((iL < iNodes) && (NULL != fgets(zcScanline, MAXLINE, pfIn))){

    if (NULL == ((*ppcSeqLabels_p)[iL] = (char *)malloc(strlen(zcScanline)))){
      sprintf(zcError,"%s:%s:%d: could not malloc %ld char* for seq label %d\n",
	      __FUNCTION__, __FILE__, __LINE__, strlen(zcScanline)+1, iL);
      return FAILURE;
    }
    strtok(zcScanline, " \040\t\n");
    strcpy((*ppcSeqLabels_p)[iL], zcScanline);

    iL++;

  } /* !EOF */

  fclose(pfIn); pfIn = NULL;

  if (iL < iNodes){
    sprintf(zcError, "%s:%s:%d: not enough labels in file %s (%d/%d)\n",
	    __FUNCTION__, __FILE__, __LINE__, zcFilename, iL, iNodes);
    return FAILURE;
  }

  /* @= */
  return OK;

} /* this is theend of ReadSequenceLabels() */


/* @* OpenFiles */
int OpenFiles(FILE **pfBal_p, FILE **pfImb_p, FILE **pfIm2_p, FILE **pfIbb_p, 
	      char zcFilename[], 
	      char zcAux[], char zcError[]){

  char *pcAux = NULL;

  /* @<open balanced file> */
  strcpy(zcAux, zcFilename);
  pcAux = strrchr(zcAux, '.');
  if (NULL != pcAux){
    *pcAux = '\0';
  }
  strcat(zcAux, "-bal.dnd");
  if (NULL == (*pfBal_p = fopen(zcAux, "w"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for writing\n", 
	    __FUNCTION__, __FILE__, __LINE__, zcAux);
    return FAILURE;
  }


  /* @<open un-balanced file> */
  strcpy(zcAux, zcFilename);
  pcAux = strrchr(zcAux, '.');
  if (NULL != pcAux){
    *pcAux = '\0';
  }
  strcat(zcAux, "-imb.dnd");
  if (NULL == (*pfImb_p = fopen(zcAux, "w"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for writing\n", 
	    __FUNCTION__, __FILE__, __LINE__, zcAux);
    return FAILURE;
  }


  /* @<open 2-un-balanced file> */
  strcpy(zcAux, zcFilename);
  pcAux = strrchr(zcAux, '.');
  if (NULL != pcAux){
    *pcAux = '\0';
  }
  strcat(zcAux, "-im2.dnd");
  if (NULL == (*pfIm2_p = fopen(zcAux, "w"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for writing\n", 
	    __FUNCTION__, __FILE__, __LINE__, zcAux);
    return FAILURE;
  }


  /* @<open un-balanced combination of balanced file> */
  strcpy(zcAux, zcFilename);
  pcAux = strrchr(zcAux, '.');
  if (NULL != pcAux){
    *pcAux = '\0';
  }
  strcat(zcAux, "-ibb.dnd");
  if (NULL == (*pfIbb_p = fopen(zcAux, "w"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for writing\n", 
	    __FUNCTION__, __FILE__, __LINE__, zcAux);
    return FAILURE;
  }


  /* @= */
  return OK;

} /* this is the end of OpenFiles() */


/* @* CreateTwoUnBalanced*/
int CreateTwoUnBalanced(int iD, int iNodes, FILE *pfOut, char **ppcSeqLabels, 
			char zcAux[], char zcError[]){

  if (NULL == pfOut){
    sprintf(zcError, "%s:%s:%d: ptr for 2-un-balanced file = NULL\n",
            __FUNCTION__, __FILE__, __LINE__);
    return FAILURE;
  }

  fprintf(pfOut, "(\n");
  CreateUnBalanced(iD, iNodes/2, pfOut, ppcSeqLabels, zcAux, zcError);
  iNodeWritten_GLOBAL++;
  fprintf(pfOut, ":%f\n", SOME_DISTANCE);
  fprintf(pfOut, ",\n");
  CreateUnBalanced(iD, iNodes/2, pfOut, ppcSeqLabels, zcAux, zcError);
  fprintf(pfOut, ":%f\n", SOME_DISTANCE);
  fprintf(pfOut, ");\n");



  /* @= */
  return OK;

} /* this is the end of CreateTwoUnBalanced() */


/* @* CreateUnBalancedBalanced */
int CreateUnBalancedBalanced(int iD, int iDepth, int iNodes, FILE *pfOut, 
			     char **ppcSeqLabels,
			     char zcAux[], char zcError[]){

  int iBalDep = ilog2(iNodes);

  if (NULL == pfOut){
    sprintf(zcError, "%s:%s:%d: ptr for un-balanced file = NULL\n",
            __FUNCTION__, __FILE__, __LINE__);
    return FAILURE;
  }

  fprintf(pfOut, "(\n");
  CreateBalanced(1, iBalDep, pfOut, ppcSeqLabels, zcAux, zcError);
  fprintf(pfOut, ":%f\n", (iNodes-iD)*SOME_DISTANCE);

  fprintf(pfOut, ",\n");
  if (iD != iDepth-1){
    CreateUnBalancedBalanced(iD+1, iDepth, iNodes, pfOut, ppcSeqLabels, 
			     zcAux, zcError);
  }
  else{
    CreateBalanced(1, iBalDep, pfOut, ppcSeqLabels, zcAux, zcError);
    fprintf(pfOut, ":%f\n", (iNodes-iD)*SOME_DISTANCE);
  }
  fprintf(pfOut, ")");
  if (iD > 1){
    fprintf(pfOut, ":%f\n", SOME_DISTANCE);
  }

  /* @= */
  return OK;

} /* this is the end of CreateUnBalancedBalanced() */



/* @* CreateUnBalanced*/
int CreateUnBalanced(int iD, int iNodes, FILE *pfOut, char **ppcSeqLabels, 
		     char zcAux[], char zcError[]){

  if (NULL == pfOut){
    sprintf(zcError, "%s:%s:%d: ptr for un-balanced file = NULL\n", 
	    __FUNCTION__, __FILE__, __LINE__);
    return FAILURE;
  }

  fprintf(pfOut, "(\n");
  sprintf(zcAux, "%s", ppcSeqLabels[iNodeWritten_GLOBAL]);
  iNodeWritten_GLOBAL++;
  fprintf(pfOut, "%s:%f\n", zcAux, (iNodes-iD)*SOME_DISTANCE);

  fprintf(pfOut, ",\n");
  if (iD != iNodes-1){
    CreateUnBalanced(iD+1, iNodes, pfOut, ppcSeqLabels, zcAux, zcError);
  }
  else{
    sprintf(zcAux, "%s", ppcSeqLabels[iNodeWritten_GLOBAL]);
    fprintf(pfOut, "%s:%f\n", zcAux, (iNodes-iD)*SOME_DISTANCE);
  }
  fprintf(pfOut, ")");
  if (iD > 1){
    fprintf(pfOut, ":%f\n", SOME_DISTANCE);
  }

  /* @= */
  return OK;

} /* this is the end of CreateUnBalanced() */




/* @* CreateBalanced */
int CreateBalanced(int iD, int iDepth, FILE *pfOut, char **ppcSeqLabels, 
		   char zcAux[], char zcError[]){


  if (NULL == pfOut){
    sprintf(zcError, "%s:%s:%d: ptr for balanced file = NULL\n", 
	    __FUNCTION__, __FILE__, __LINE__);
    return FAILURE;
  }

  fprintf(pfOut, "(\n");
  if (iD == iDepth){
    sprintf(zcAux, "%s", ppcSeqLabels[iNodeWritten_GLOBAL]);
    iNodeWritten_GLOBAL++;
    fprintf(pfOut, "%s:%f\n", zcAux, SOME_DISTANCE);
  }
  else {
    CreateBalanced(iD+1, iDepth, pfOut, ppcSeqLabels, zcAux, zcError);
  }
  fprintf(pfOut, ",\n");
  if (iD == iDepth){
    sprintf(zcAux, "%s", ppcSeqLabels[iNodeWritten_GLOBAL]);
    iNodeWritten_GLOBAL++;
    fprintf(pfOut, "%s:%f\n", zcAux, SOME_DISTANCE);
  }
  else {
    CreateBalanced(iD+1, iDepth, pfOut, ppcSeqLabels, zcAux, zcError);
  }
  fprintf(pfOut, ")");
  if (iD > 1){
    fprintf(pfOut, ":%f", SOME_DISTANCE);
  }
  fprintf(pfOut, "\n");

  /* @= */
  return OK;

} /* this is the end of CreateBalanced() */


/* @* Ipow */
int Ipow(int *iNodes_p, int iDepth, char zcError[]){

  int i;

  *iNodes_p = UNITY;

  for (i = 0; i < iDepth; i++){
    *iNodes_p *= BASE2;
  }

  /* @= */
  return OK;

} /* this is the end of Ipow() */

/* @* Commandline */
enum {ARG_EXEC = 0, ARG_DEPTH, ARG_INFILE, ARGC};
int Commandline(int *iDepth_p, char zcFilename[], 
		int argc, char *argv[], char zcError[]){

  /* @<variables local to Commandline@> */
  char *pcEnd = NULL;
  char zcUsage[] = 
    "Example: ./constructTree   5    mylabels.txt\n"
    "Usage:   <exec>         <depth> <labels>\n"
    "          |                |       |\n"
    "          |                |       +-file with sequence labels\n"
    "          |                +-depth of tree (#nodes=2^depth)\n"
    "          +-executable\n";

  /* @<check number of arguments@> */
  if (ARGC != argc){
    sprintf(zcError, "%s:%d: %d args expected, %d supplied\n%s", 
	    __FILE__, __LINE__, ARGC-1, argc-1, zcUsage);
    return FAILURE;
  }

  /* @<convert integer@> */
  *iDepth_p = (int)strtol(argv[ARG_DEPTH], &pcEnd, BASE10);
  if ('\0' != *pcEnd){
    sprintf(zcError, "%s:%d: depth not clean int (%s)\n%s", 
	    __FILE__, __LINE__, argv[ARG_DEPTH], zcUsage);
    return FAILURE;
  }

  /* @<get filename for input file> */
  strcpy(zcFilename, argv[ARG_INFILE]);


  /* @<convert double@> */
  //  *d_p = (double)strtod(argv[ARG_D], &pcEnd, BASE10);
  //if ('\0' != *pcEnd){
  //  sprintf(zcError, "%s:%d:  not clean double (%s)\n%s",
  //          __FILE__, __LINE__, argv[ARG_D], zcUsage);
  //  return FAILURE;
  //}


  /* @= */
  return OK;

} /* this is the end of Commandline() */


/* @* ErrorExit */
void ErrorExit(char zcError[]){
  printf("+-----------------------------------+\n"
	 "| Program did NOT run to completion |\n"
	 "+-----------------------------------+\n"
         "%s",
         zcError);
  exit(FAILURE);
} /* this is the end of ErrorExit() */




/* @i
 * EOF 
 */
